NPCMuter = {
	name = "NPCMuter",
	author = "Terrillyn",
	version = "100016.002",
	CurSV = {},
	AWSV = {},
	CharSV = {},
}

local VOLUMESTATE = {
	UNMUTED = "unmuted",
	UNKNOWN = -1,
	MUTED = "muted",
	CONVO = "convo",
}

local VOLUMECHANNEL = {
	VOICE = { name="voice", setting=AUDIO_SETTING_VO_VOLUME },
	AMBIENT = { name="ambient", setting=AUDIO_SETTING_AMBIENT_VOLUME },
}

local CONVOTYPE = {
	BANKER = "Bank",
	GUILDSTORE = "Guild Store",
	STORE = "Store",
	UNKNOWN = "",
}


NPCMuter.session = {
	delayed_calls = {},
	vol_state = {
		voice = VOLUMESTATE.UNKNOWN,
		ambient = VOLUMESTATE.UNKNOWN,
	},
	old_levels = {
		voice = -1,
		ambient = -1,
	},
	player_state = {
		lastConvoType = "",
		banker = false,
		store = false,
		guildstore = false,
		heard = false,
		blacklisted = false,
		convo = false,
	},
}

local defaults = {
	internal_version = "",
	accountwide = false,
	firstrun = true,
	levels = {
		ambient = {
			muted = -1,
			unmuted = -1,
			convo = -1,
		},
		voice = {
			muted = -1,
			unmuted = -1,
			convo = -1,
		},
	},
	options = {
		bankers = false,
		stores = false,
		guildstores = false,
		heard = false,
		exclusivemode = false, -- unimplemented
		mute_length_mod = 1,
	},
	blacklists = {
		npcs = {},
		lines = {},
	},
}

function NPCMuter.OnAddOnLoaded(event, addonName)
	if addonName ~= NPCMuter.name then return end

	NPCMuter.AWSV = ZO_SavedVars:NewAccountWide("NPCMuter_Settings", "1", nil, defaults)
	NPCMuter.CharSV = ZO_SavedVars:New("NPCMuter_Settings", "1", nil, defaults)

	if NPCMuter.AWSV.accountwide then
		NPCMuter.CurSV = NPCMuter.AWSV
		NPCMuter.CharSV.version = nil
		NPCMuter.CharSV = ZO_SavedVars:New("NPCMuter_Settings", "1", nil, NPCMuter_Settings['Default'][GetDisplayName()]['$AccountWide'])
	else
		NPCMuter.CurSV = NPCMuter.CharSV
	end

	if NPCMuter.CurSV.firstrun then
		NPCMuter.CurSV.firstrun = false
		NPCMuter.CurSV.levels.voice.unmuted = tonumber(GetSetting(SETTING_TYPE_AUDIO, AUDIO_SETTING_VO_VOLUME))
		NPCMuter.CurSV.levels.ambient.unmuted = tonumber(GetSetting(SETTING_TYPE_AUDIO, AUDIO_SETTING_AMBIENT_VOLUME))

		NPCMuter.CurSV.levels.voice.muted = 0
		NPCMuter.CurSV.levels.voice.convo = NPCMuter.CurSV.levels.voice.unmuted / 2

		NPCMuter.CurSV.levels.ambient.muted = 0
		NPCMuter.CurSV.levels.ambient.convo = NPCMuter.CurSV.levels.ambient.unmuted / 2
		NPCMuter.CurSV.internal_version = "rewrite1"
	else
		NPCMuter.UpdateSavedVars()
	end
	NPCMuter.UpdateSavedVars()
	NPCMuter.CreateSlashCommands()
	NPCMuter.CreateSettingsMenu()

	EVENT_MANAGER:UnregisterForEvent(NPCMuter.name, EVENT_ADD_ON_LOADED)
	EVENT_MANAGER:RegisterForEvent(NPCMuter.name, EVENT_PLAYER_ACTIVATED, NPCMuter.OnPlayerActivated)
end

function NPCMuter.OnPlayerActivated()
	NPCMuter.session.old_levels.voice = tonumber(GetSetting(SETTING_TYPE_AUDIO, AUDIO_SETTING_VO_VOLUME))
	NPCMuter.session.old_levels.ambient = tonumber(GetSetting(SETTING_TYPE_AUDIO, AUDIO_SETTING_AMBIENT_VOLUME))

	local function restoreVolumes()
		SetSetting(SETTING_TYPE_AUDIO, VOLUMECHANNEL.VOICE.setting, NPCMuter.session.old_levels.voice)
		SetSetting(SETTING_TYPE_AUDIO, VOLUMECHANNEL.AMBIENT.setting, NPCMuter.session.old_levels.ambient)
	end
	
	ZO_PreHook("ReloadUI", restoreVolumes)
	ZO_PreHook("Logout", restoreVolumes)
	ZO_PreHook("Quit", restoreVolumes)

	Terril_lib.IsAddonOutOfDate(NPCMuter, true)
	NPCMuter.UpdateVolumeStates()

	EVENT_MANAGER:RegisterForEvent(NPCMuter.name, EVENT_CHATTER_BEGIN, NPCMuter.EVENT_CHATTER_BEGIN)
	EVENT_MANAGER:RegisterForEvent(NPCMuter.name, EVENT_CHATTER_END, NPCMuter.EVENT_CHATTER_END)
	EVENT_MANAGER:RegisterForEvent(NPCMuter.name, EVENT_CHAT_MESSAGE_CHANNEL, NPCMuter.EVENT_CHAT_MESSAGE_CHANNEL)
	EVENT_MANAGER:UnregisterForEvent(NPCMuter.name, EVENT_PLAYER_ACTIVATED)
end

function NPCMuter.SetVolumeState(channel, state)
	--d("NPCMuter: Setting " .. channel.name .. " Volume to " .. state .. "[" .. tostring(NPCMuter.CurSV.levels[channel.name][state]) .. "%]")
	SetSetting(SETTING_TYPE_AUDIO, channel.setting, NPCMuter.CurSV.levels[channel.name][state])
end

function NPCMuter.UpdateVolumeStates()
	local playerState = NPCMuter.session.player_state
	local volumeStates = NPCMuter.session.vol_state

	local function SetAll(state)
		for _,ch in pairs(VOLUMECHANNEL) do
			volumeStates[ch.name] = state
			NPCMuter.SetVolumeState(ch, state)
		end
	end

	if NPCMuter.CurSV.options.bankers and playerState.banker then
		SetAll(VOLUMESTATE.MUTED)
	elseif NPCMuter.CurSV.options.stores and playerState.store then
		SetAll(VOLUMESTATE.MUTED)
	elseif NPCMuter.CurSV.options.guildstores and playerState.guildstore then
		SetAll(VOLUMESTATE.MUTED)
	elseif  playerState.convo then
		SetAll(VOLUMESTATE.CONVO)
	elseif playerState.heard then
		SetAll(VOLUMESTATE.MUTED)
	elseif playerState.blacklisted then
		SetAll(VOLUMESTATE.MUTED)
	else
		if NPCMuter.CurSV.options.exclusivemode then SetAll(VOLUMESTATE.UNMUTED)
		else SetAll(VOLUMESTATE.MUTED) end
	end
end

function NPCMuter.DelayCall(name, delay_ms, func)
	local unique_name = NPCMuter.name .. "." .. name

	if NPCMuter.session.delayed_calls[unique_name] then EVENT_MANAGER:UnregisterForUpdate(unique_name) end
	NPCMuter.session.delayed_calls[unique_name] = true

	EVENT_MANAGER:RegisterForUpdate(unique_name, delay_ms,
		function()
			EVENT_MANAGER:UnregisterForUpdate(unique_name)
			NPCMuter.session.delayed_calls[unique_name] = nil
			func()
		end)
end

function NPCMuter.IsPlayerName(name) -- fix
	return string.match(name, '%^[MF]x$') ~= nil
end

function NPCMuter.TextToTime(text)
	return math.floor(((string.len(text) / NPCMuter.CurSV.options.mute_length_mod) + 2) * 1000)
end

function NPCMuter.EVENT_CHAT_MESSAGE_CHANNEL(eventCode, messageType, fromName, text, isCustomerService, displayName)
	if (messageType ~= CHAT_CHANNEL_MONSTER_SAY and messageType ~= CHAT_CHANNEL_MONSTER_YELL) then return end
	if isCustomerService or NPCMuter.session.player_state.convo then return end
	if NPCMuter.IsPlayerName(fromName) then return end
	local npcName = string.gsub(fromName, '%^.*$', '')
	--d(string.format("fromname: %s, newName: %s", fromName, npcName))

	if NPCMuter.CurSV.options.heard then
		NPCMuter.session.player_state.heard = false

		local short = Terril_lib.firstLetters(text)
		if not NPCMuter.CurSV.blacklists.lines[npcName] then NPCMuter.CurSV.blacklists.lines[npcName] = {} end
		if NPCMuter.CurSV.blacklists.lines[npcName][text] then
			NPCMuter.session.player_state.heard = true
		else
			NPCMuter.CurSV.blacklists.lines[npcName][text] = true
		end
	end

	NPCMuter.session.player_state.blacklisted = false
	if NPCMuter.CurSV.blacklists.npcs[npcName] then
		NPCMuter.session.player_state.blacklisted = true
	end

	if NPCMuter.session.player_state.blacklisted or NPCMuter.session.player_state.heard then
		NPCMuter.DelayCall("mute_delay", NPCMuter.TextToTime(text), function()
			NPCMuter.session.player_state.blacklisted = false
			NPCMuter.session.player_state.heard = false
			NPCMuter.UpdateVolumeStates()
		end)
	end

	NPCMuter.UpdateVolumeStates()
end

local CONVOTYPE = {
	CONVO = {setting="convo", id=-1},
	STORE = {setting="store", id=600},
	STABLE = {setting="stable", id=3100},
	BANK = {setting="banker", id=1200},
	GUILDBANK = {setting="guildbank", id=3300},
	GUILDSTORE = {setting="guildstore", id=3400},
	BID = {setting="bid", id=3800},
}

function NPCMuter.UpdateConvoType(optionCount)
	if optionCount and optionCount > 0 then
		for i=1, optionCount do
			local _, optionType, _, _, _ = GetChatterOption(i)

			if optionType == CONVOTYPE.STORE.id or optionType == CONVOTYPE.STABLE.id then
				NPCMuter.session.player_state[CONVOTYPE.STORE.setting] = true
				break
			elseif optionType == CONVOTYPE.BANK.id or optionType == CONVOTYPE.GUILDBANK.id then
				NPCMuter.session.player_state[CONVOTYPE.BANK.setting] = true
				break
			elseif optionType == CONVOTYPE.GUILDSTORE.id or optionType == CONVOTYPE.BID.id then
				NPCMuter.session.player_state[CONVOTYPE.GUILDSTORE.setting] = true
				break
			end
		end
	end
	NPCMuter.session.player_state[CONVOTYPE.CONVO.setting] = true
end

function NPCMuter.EVENT_CHATTER_BEGIN(eventCode, optionCount)
	NPCMuter.UpdateConvoType(optionCount)
	NPCMuter.UpdateVolumeStates()
end

function NPCMuter.EVENT_CHATTER_END(eventCode)
	local settingChanged = false
	for t,d in pairs(CONVOTYPE) do
		if NPCMuter.session.player_state[d.setting] then
			NPCMuter.session.player_state[d.setting] = false
			settingChanged = true
		end
	end

	if settingChanged then NPCMuter.UpdateVolumeStates() end
end
------------------------------------------------------------------------------------

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
------------------------------------------------------------------------------------

-- update 100013.114 savedVars to current
function NPCMuter.UpdateSavedVars()
	if (NPCMuter.CurSV.internal_version == "" or NPCMuter.CurSV.internal_version == nil) and NPCMuter.CurSV.version == "1" then
		NPCMuter.CurSV.levels.voice.unmuted = NPCMuter.CurSV.default_vo_vol
		NPCMuter.CurSV.levels.voice.muted = NPCMuter.CurSV.muted_vo_vol
		NPCMuter.CurSV.levels.voice.convo = NPCMuter.CurSV.convo_vo_vol
		NPCMuter.CurSV.levels.ambient.unmuted = NPCMuter.CurSV.default_am_vol
		NPCMuter.CurSV.levels.ambient.muted = NPCMuter.CurSV.muted_am_vol
		NPCMuter.CurSV.levels.ambient.convo = NPCMuter.CurSV.convo_am_vol

		NPCMuter.CurSV.options.bankers = NPCMuter.CurSV.mute_bankers
		NPCMuter.CurSV.options.stores = NPCMuter.CurSV.mute_stores
		NPCMuter.CurSV.options.guildstores = NPCMuter.CurSV.mute_guildstores
		NPCMuter.CurSV.options.heard = NPCMuter.CurSV.mute_heard
		NPCMuter.CurSV.options.exclusivemode = NPCMuter.CurSV.exclusive_mute_mode
		NPCMuter.CurSV.options.mute_length_mod = NPCMuter.CurSV.mute_modifier

		for i,v in ipairs(NPCMuter.CurSV.blacklist) do
			NPCMuter.CurSV.blacklists.npcs[v] = true
		end
		for name,line in pairs(NPCMuter.CurSV.heardlines) do
			NPCMuter.CurSV.blacklists.lines[name] = line
		end

		NPCMuter.CurSV.default_vo_vol = nil
		NPCMuter.CurSV.muted_vo_vol = nil
		NPCMuter.CurSV.convo_vo_vol = nil
		NPCMuter.CurSV.default_am_vol = nil
		NPCMuter.CurSV.muted_am_vol = nil
		NPCMuter.CurSV.convo_am_vol = nil

		NPCMuter.CurSV.mute_bankers = nil
		NPCMuter.CurSV.mute_stores = nil
		NPCMuter.CurSV.mute_guildstores = nil
		NPCMuter.CurSV.mute_heard = nil
		NPCMuter.CurSV.exclusive_mute_mode = nil
		NPCMuter.CurSV.mute_modifier = nil

		NPCMuter.CurSV.blacklist = nil
		NPCMuter.CurSV.heardlines = nil

		NPCMuter.CurSV.world_am_vol = nil
	end

	NPCMuter.CurSV.internal_version = "rewrite1"
end


-------------------------------------------------------------------------------------------------
function NPCMuter.CreateSlashCommands()

	local ProcessTarget = function(unitName, listFunc)
		if unitName ~= "" then
			listFunc(unitName)
		else
			EVENT_MANAGER:RegisterForEvent(NPCMuter.name, EVENT_RETICLE_TARGET_CHANGED, function()
				local name = GetRawUnitName("reticleover")
				if not NPCMuter.IsPlayerName(name) then listFunc(GetUnitName("reticleover")) end
				EVENT_MANAGER:UnregisterForEvent(NPCMuter.name, EVENT_RETICLE_TARGET_CHANGED)
			end)
		end
	end

	SLASH_COMMANDS["/mute"] = function(arg) ProcessTarget(arg, function(name)
		NPCMuter.CurSV.blacklists.npcs[name] = true
		d("NPCMuter: " .. name .. " added to list.")
	end) end
	SLASH_COMMANDS["/unmute"] = function(arg) ProcessTarget(arg, function(name)
		NPCMuter.CurSV.blacklists.npcs[name] = nil
		d("NPCMuter: " .. name .. " removed from list.")
	end) end
	SLASH_COMMANDS["/npcmuter"] = function(extra)
		if extra == "help" then
			d("NPCMuter CommandList: ")
			d("'/npcmuter help' - shows this list.")
			d("'/npcmuter dump' - prints out all settings.")
			d("'/npcmuter clearlist' - clears mutelist.")
			d("'/npcmuter clearheard' - clears heardlines list.")
			--d("'/npcmuter debug' - toggles debug output.")
			d("'/npcmuter list' - prints out mutelist.")
			d("'/mute <name>' - mutes named npc.")
			d("'/unmute <name>' - unmutes named npc.")
		elseif extra == "clearlist" then
			NPCMuter.CurSV.blacklists.npcs = {}
			d("Mutelist cleared.")
		elseif extra == "clearheard" then
			NPCMuter.CurSV.blacklists.lines = {}
			d("Heardlines cleared.")
		elseif extra == "dump" then
			d("Addon Dump:")
			d("-    " .. NPCMuter.name .. " v" .. NPCMuter.version)
			d("-    Debug Mode: " .. "disabled_unimplemented")
			local function listTable(t, prefix)
				prefix = prefix or ""
				for k,v in pairs(t) do
					d(prefix .. tostring(k) .. ": " .. tostring(v))
				end
			end
			d("accountwide: " .. tostring(NPCMuter.CurSV.accountwide))
			listTable(NPCMuter.CurSV.levels.voice, "voice.")
			listTable(NPCMuter.CurSV.levels.ambient, "ambient.")
			listTable(NPCMuter.CurSV.options, "options.")
		-- elseif extra == "debug" then
		-- 	NPCMuter.debug_mode = not NPCMuter.debug_mode
		-- 	d("NPCMuter.debug_mode set to " .. tostring(NPCMuter.debug_mode))
		elseif extra == "list" then
			d("Currently muted npcs:")
			for k,v in pairs(NPCMuter.CurSV.blacklists.npcs) do
				d("-    " .. k)
			end
		elseif extra == "" then
			LAM2:OpenToPanel(panel_ref)
		end
	end
end

function NPCMuter.CreateSettingsMenu()
	local LAM2 = LibStub("LibAddonMenu-2.0")

	--settings panel
	local panelData = {
		type = "panel",
		name = NPCMuter.name,
		author = NPCMuter.author,
		version = NPCMuter.version,
	}
	local panel_ref = LAM2:RegisterAddonPanel(NPCMuter.name .. "OptionsPanel", panelData)

	local optionsData = {
		{
			type = "checkbox",
			name = "Account-Wide",
			getFunc = function()
				return NPCMuter.AWSV.accountwide
			end,
			setFunc = function(val)
				if val then
					--load
					NPCMuter.AWSV.version = nil
					NPCMuter.AWSV = ZO_SavedVars:NewAccountWide("NPCMuter_Settings", "1", nil, NPCMuter_Settings['Default'][GetDisplayName()][GetUnitName('player')])
					NPCMuter.CurSV = NPCMuter.AWSV
				else
					NPCMuter.CharSV.version = nil
					NPCMuter.CharSV = ZO_SavedVars:New("NPCMuter_Settings", "1", nil, NPCMuter_Settings['Default'][GetDisplayName()]['$AccountWide'])
					NPCMuter.CurSV = NPCMuter.CharSV
				end

				NPCMuter.AWSV.accountwide = val
			end,
		},
		{
			type = "header",
			name = "Voice-Over Volumes",
			width = "half",
			registerForRefresh = true,
		},
		{
			type = "slider",
			name = "Unmuted Dialogue Volume",
			tooltip = "The default volume for voice-overs when not muted.",
			min = 0,
			max = 100,
			getFunc = function() return NPCMuter.CurSV.levels.voice.unmuted end,
			setFunc = function(val)
				NPCMuter.CurSV.levels.voice.unmuted = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "slider",
			name = "Muted Dialogue Volume",
			tooltip = "The volume of muted voice-overs, default zero to mute.",
			min = 0,
			max = 100,
			getFunc = function() return NPCMuter.CurSV.levels.voice.muted end,
			setFunc = function(val) NPCMuter.CurSV.levels.voice.muted = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "slider",
			name = "Conversation Dialogue Volume",
			tooltip = "The volume of conversation voice-overs. (in conversation windows)",
			min = 0,
			max = 100,
			getFunc = function() return NPCMuter.CurSV.levels.voice.convo end,
			setFunc = function(val) NPCMuter.CurSV.levels.voice.convo = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "checkbox",
			name = "Mute bankers",
			tooltip = "Enable this to keep banker windows muted.",
			getFunc = function() return NPCMuter.CurSV.options.bankers end,
			setFunc = function(val) NPCMuter.CurSV.options.bankers = val end,
		},
		{
			type = "checkbox",
			name = "Mute store npcs",
			tooltip = "Enable this to keep store npc windows muted.",
			getFunc = function() return NPCMuter.CurSV.options.stores end,
			setFunc = function(val) NPCMuter.CurSV.options.stores = val end,
		},
		{
			type = "checkbox",
			name = "Mute guild store npcs",
			tooltip = "Enable this to keep guild store npc windows muted.",
			getFunc = function() return NPCMuter.CurSV.options.guildstores end,
			setFunc = function(val) NPCMuter.CurSV.options.guildstores = val end,
		},
		{
			type = "checkbox",
			name = "Mute heard lines",
			tooltip = "Enable this to automatically mute any lines you've already heard from npcs.",
			getFunc = function() return NPCMuter.CurSV.options.heard end,
			setFunc = function(val) NPCMuter.CurSV.options.heard = val end,
		},
		{
			type = "checkbox",
			name = "Exclusive mute mode",
			width = "full",
			tooltip = "The voice volume state becomes unmuted by default only muting when required (such as when a blacklisted npc talks).",
			getFunc = function() return NPCMuter.CurSV.options.exclusivemode end,
			setFunc = function(val)
				NPCMuter.CurSV.options.exclusivemode = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "header",
			name = "Ambient Volumes",
			width = "half",
		},
		{
			type = "slider",
			name = "Unmuted Ambient Volume",
			tooltip = "The Ambient volume level when a world npc is talking. (outside of conversations)",
			min = 0,
			max = 100,
			getFunc = function() return NPCMuter.CurSV.levels.ambient.unmuted end,
			setFunc = function(val) NPCMuter.CurSV.levels.ambient.unmuted = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "slider",
			name = "Muted Ambient Volume",
			tooltip = "The Ambient volume level when not talking to an npc.",
			min = 0,
			max = 100,
			getFunc = function() return NPCMuter.CurSV.levels.ambient.muted end,
			setFunc = function(val)
				NPCMuter.CurSV.levels.ambient.muted = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "slider",
			name = "Conversation Ambient Volume",
			tooltip = "The Ambient volume level when talking to an npc.",
			min = 0,
			max = 100,
			getFunc = function() return NPCMuter.CurSV.levels.ambient.convo end,
			setFunc = function(val) NPCMuter.CurSV.levels.ambient.convo = val
				NPCMuter.UpdateVolumeStates()
			end,
		},
		{
			type = "header",
			name = "Miscellaneous",
			width = "half",
		},
		{
			type = "slider",
			name = "Unmute length modifier",
			tooltip = "How long an npc will remain unmuted when they speak. (smaller value = longer) (recommended 13.5); When in exclusive mute mode, this will adjust the mute length instead.",
			step = 0.5,
			min = 10,
			max = 20,
			getFunc = function() return NPCMuter.CurSV.options.mute_length_mod end,
			setFunc = function(val) NPCMuter.CurSV.options.mute_length_mod = val end,
		},
		-- {
		-- 	type = "checkbox",
		-- 	name = "Enable debug messages",
		-- 	width = "half",
		-- 	tooltip = "Makes the addon more verbose.",
		-- 	getFunc = function() return NPCMuter.debug_mode end,
		-- 	setFunc = function(val) NPCMuter.debug_mode = val end,
		-- },
		-- {
		-- 	type = "button",
		-- 	name = "Reset Volumes",
		-- 	tooltip = "This will reset volume settings to default.",
		-- 	width = "half",
		-- 	func = function()
		-- 		settings.muted_vo_vol = 0
		-- 		settings.convo_vo_vol = settings.default_vo_vol or 50
		-- 		settings.muted_am_vol = settings.default_am_vol or 50
		-- 		settings.convo_am_vol = (settings.default_am_vol/2) or 25
		-- 		ReloadUI()
		-- 		d("NPCMuter settings reset.")
		-- 	end
		-- },
		{
			type = "button",
			name = "Clear mutelist",
			tooltip = "This will unmute all muted npcs.",
			width = "half",
			func = function()
				NPCMuter.CurSV.blacklists.npcs = {}
			end
		},
		{
			type = "button",
			name = "Clear heard lines",
			tooltip = "This will reset the heard lines.",
			width = "half",
			func = function()
				NPCMuter.CurSV.blacklists.lines = {}
			end
		},
	}
	LAM2:RegisterOptionControls(NPCMuter.name .. "OptionsPanel", optionsData)
end
-------------------------------------------------------------------------------------------------
EVENT_MANAGER:RegisterForEvent(NPCMuter.name, EVENT_ADD_ON_LOADED, NPCMuter.OnAddOnLoaded)